package be.iminds.ilabt.jfed.experimenter_gui.renew_report;

import be.iminds.ilabt.jfed.experiment.ExperimentController;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experimenter_gui.dialogs.RenewSliceDialog;
import be.iminds.ilabt.jfed.experimenter_gui.ui.AbstractDialogFactory;
import javafx.application.Platform;
import javafx.fxml.FXMLLoader;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import javax.inject.Named;
import javax.inject.Provider;
import javax.inject.Singleton;
import java.time.Instant;

@Singleton
public class RenewReportDialogFactory extends AbstractDialogFactory {
    private static final Logger LOG = LoggerFactory.getLogger(RenewReportDialogFactory.class);

    @Inject
    public RenewReportDialogFactory(Provider<FXMLLoader> loaderProvider,
                                    @Named("parentStage") Provider<Stage> parentStageProvider) {
        super(loaderProvider, parentStageProvider );
    }

    @Override
    protected String getTitle() {
        return "Renew Report";
    }

    //job succeeded, but calls might still have failed: check experimentTaskStatus
    public void showDialog(@Nonnull Instant requestedRenewInstant,
                           @Nonnull ExperimentController experimentController,
                           @Nullable ExperimentTaskStatus experimentTaskStatus) {
        try {
            assert Platform.isFxApplicationThread();

            StageAndController stageAndController = createStage();
            RenewReportDialog controller = (RenewReportDialog) stageAndController.getController();

            controller.setRequestedRenewInstant(requestedRenewInstant);
            controller.setExperimentController(experimentController);
            controller.setJobSuccessExperimentTaskStatus(experimentTaskStatus);

            controller.calculateStatus();

            stageAndController.getStage().showAndWait();
        } catch (Throwable t) {
            LOG.error("Failed to show Renew Report", t);
            throw t;
        }
    }

    //job failed
    public void showDialog(@Nonnull Instant requestedRenewInstant,
                           @Nonnull ExperimentController experimentController,
                           @Nonnull Throwable t) {
        StageAndController stageAndController = createStage();
        RenewReportDialog controller = (RenewReportDialog) stageAndController.getController();

        controller.setRequestedRenewInstant(requestedRenewInstant);
        controller.setExperimentController(experimentController);
        controller.setJobFailureThrowable(t);

        controller.calculateStatus();

        stageAndController.getStage().showAndWait();
    }
}
